'use strict'

// var fs = require('mz/fs')
// var cp = require('mz/child_process')
// var Promise = require('bluebird')
// var mkdirp = require('mkdirp-then')
var path = require('path')
var safe = require('safetimeout')
var LE = require('greenlock')
var LE_STORE = require('le-store-certbot')
var LE_HTTP_CHALLENGE = require('le-challenge-fs')
var LE_SNI_CHALLENGE = require('le-challenge-sni')
var CONSTANTS = require('./constants')
var STATIC_PATH = CONSTANTS.STATIC_PATH
var lock = {}
module.exports = load

/**
 * Attempts to register a hostname with letsencrypt.
 *
 * @param {Object} opts
 * @return {Promise}
 */
function load (domains, opts) {
  var key = domains.join()
  // Lock to ensure we are not fetching a host multiple times.
  if (lock[key]) return lock[key]
  // var stdio = opts.debug ? 'pipe' : 'ignore'
  var configDir = opts.dir || path.join(process.env.HOME, '/letsencrypt/etc')
  var store = LE_STORE.create({
    configDir: configDir,
    webrootPath: STATIC_PATH,
    debug: opts.debug
  })
  var httpChallenge = LE_HTTP_CHALLENGE.create({
    webrootPath: STATIC_PATH,
    loopbackPort: opts.ports.http,
    debug: opts.debug
  })
  var sniChallenge = LE_SNI_CHALLENGE.create({
    debug: opts.debug
  })
  var le = LE.create({
    debug: opts.debug,
    server: opts.debug ? LE.stagingServerUrl : LE.productionServerUrl,
    challenges: {
      'http-01': httpChallenge,
      'tls-sni-01': sniChallenge,
      'tls-sni-02': sniChallenge
    },
    challengeType: 'http-01',
    store: store
  })

  var hostLock = le.check({ domains: domains })
    .then(checkExpiredOrMissing)
    .then(extractCredentials)
    .catch(renewOrRegister)
    .then(function (data) {
      console.log('SUCCESS', data)
    })
    .catch(function (err) {
      console.log('ERROR', err.description)
    })

  function checkExpiredOrMissing (data) {
    // Check for expired or non existing certs.
    if (!data || data.expiresAt <= 0) throw new Error('Certificate expired.')
    return data
  }

  function extractCredentials (data) {
    var key = data.privkey
    var cert = data.chain
    // Remove cert later.
    safe.setTimeout(renewOrRegister, data.expiresAt - 8.64e+8)
    // Return a useful cert for SNI.
    return { key: key, cert: cert }
  }

  function renewOrRegister () {
    return le.register({
      domains: domains,
      email: opts.email,
      agreeTos: opts.agreeTos,
      challengeType: 'http-01',
      rsaKeySize: 2048
    }).then(extractCredentials)
  }

  lock[key] = hostLock
  return hostLock
}
